/*
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
 * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// local
#include "application.h"
#include "application_manager.h"
#include "desktopfilereader.h"
#include "taskcontroller.h"

// unity-mir
#include "logging.h"

// mir
#include <mir/scene/session.h>
#include <mir/scene/snapshot.h>
#include <mir/scene/prompt_session.h>
#include <mir/scene/prompt_session_manager.h>

namespace ms = mir::scene;

namespace unitymir
{

QMutex screenshotMutex;

Application::Application(const QSharedPointer<TaskController>& taskController,
                         DesktopFileReader *desktopFileReader,
                         State state,
                         const QStringList &arguments,
                         const std::shared_ptr<ms::PromptSessionManager>& promptSessionManager,
                         QObject *parent)
    : ApplicationInfoInterface(desktopFileReader->appId(), parent)
    , m_taskController(taskController)
    , m_desktopData(desktopFileReader)
    , m_pid(0)
    , m_stage((m_desktopData->stageHint() == "SideStage") ? Application::SideStage : Application::MainStage)
    , m_state(state)
    , m_focused(false)
    , m_canBeResumed(true)
    , m_fullscreen(false)
    , m_visible(false)
    , m_arguments(arguments)
    , m_suspendTimer(new QTimer(this))
    , m_promptSessionManager(promptSessionManager)
    , m_screenShotGuard(new Guard)
{
    DLOG("Application::Application (this=%p, appId=%s, state=%d", this, qPrintable(desktopFileReader->appId()),
         static_cast<int>(state));

    m_suspendTimer->setSingleShot(true);
    connect(m_suspendTimer, SIGNAL(timeout()), this, SLOT(suspend()));

    // FIXME(greyback) need to save long appId internally until upstart-app-launch can hide it from us
    m_longAppId = desktopFileReader->file().remove(QRegExp(".desktop$")).split('/').last();

    connect(this, &Application::foregroundSessionChanged, this, &Application::updateScreenshot);
}

Application::~Application()
{
    {
        // In case we get a threaded screenshot callback once the application is deleted.
        QMutexLocker lk(&screenshotMutex);
        m_screenShotGuard.clear();
    }

    DLOG("Application::~Application (this=%p)", this);
    delete m_desktopData;
}

bool Application::isValid() const
{
    return m_desktopData->loaded();
}

QString Application::desktopFile() const
{
    return m_desktopData->file();
}

QString Application::appId() const
{
    return m_desktopData->appId();
}

QString Application::name() const
{
    return m_desktopData->name();
}

QString Application::comment() const
{
    return m_desktopData->comment();
}

QUrl Application::icon() const
{
    QString iconString = m_desktopData->icon();
    QString pathString = m_desktopData->path();

    if (QFileInfo(iconString).exists()) {
        return QUrl(iconString);
    } else if (QFileInfo(pathString + '/' + iconString).exists()) {
        return QUrl(pathString + '/' + iconString);
    } else {
        return QUrl("image://theme/" + iconString);
    }
}

QString Application::exec() const
{
    return m_desktopData->exec();
}

Application::Stage Application::stage() const
{
    return m_stage;
}

Application::State Application::state() const
{
    return m_state;
}

bool Application::focused() const
{
    return m_focused;
}

QUrl Application::screenshot() const
{
    return m_screenshot;
}

bool Application::fullscreen() const
{
    return m_fullscreen;
}

std::shared_ptr<ms::Session> Application::session() const
{
    return m_session;
}

std::shared_ptr<ms::Session> Application::foregroundSession() const
{
    std::shared_ptr<ms::Session> child;

    QListIterator<std::shared_ptr<ms::PromptSession>> iter(m_promptSessions);
    iter.toBack();
    while (iter.hasPrevious()) {
        std::shared_ptr<ms::PromptSession> promptSession = iter.previous();

        // get top session
        m_promptSessionManager->for_each_provider_in(promptSession,
            [&child](std::shared_ptr<ms::Session> const& provider) {
                if (provider->default_surface() != nullptr)
                    child = provider;
            });

        if (child)
            return child;
    }

    return session();
}

std::shared_ptr<ms::PromptSession> Application::activePromptSession() const
{
    if (m_promptSessions.count() > 0)
        return m_promptSessions.back();
    return nullptr;
}

bool Application::containsProcess(pid_t pid) const
{
    if (m_pid == pid)
        return true;

    QListIterator<std::shared_ptr<ms::PromptSession>> iter(m_promptSessions);
    while(iter.hasNext()) {
        std::shared_ptr<ms::PromptSession> promptSession = iter.next();

        std::shared_ptr<ms::Session> helper = m_promptSessionManager->helper_for(promptSession);
        if (helper && helper->process_id() == pid)
            return true;

        bool found = false;
        m_promptSessionManager->for_each_provider_in(promptSession,
            [&found, pid](std::shared_ptr<ms::Session> const& provider) {
                if (provider->process_id() == pid)
                    found = true;
            });
        if (found)
            return true;
    }
    return false;
}

bool Application::visible() const
{
    return m_visible;
}

bool Application::canBeResumed() const
{
    return m_canBeResumed;
}

void Application::setCanBeResumed(const bool resume)
{
    m_canBeResumed = resume;
}

pid_t Application::pid() const
{
    return m_pid;
}

void Application::setPid(pid_t pid)
{
    m_pid = pid;
}

void Application::setSession(const std::shared_ptr<ms::Session>& session)
{
    DLOG("Application::setSession (this=%p, session=%p)", this, session.get());

    // TODO(greyback) what if called with new surface?
    m_session = session;
    m_visible = true; // bit of an assumption that, but no other way to deduce an actual Surface has been created
    checkSessionChanges();
}

void Application::appendPromptSession(const std::shared_ptr<ms::PromptSession>& promptSession)
{
    DLOG("Application::setPromptSession (this=%p, promptSession=%p)", this, promptSession ? promptSession.get() : nullptr);

    m_promptSessions.append(promptSession);
    checkSessionChanges();
}

void Application::removePromptSession(const std::shared_ptr<ms::PromptSession>& promptSession)
{
    DLOG("Application::setPromptSession (this=%p, promptSession=%p)", this, promptSession ? promptSession.get() : nullptr);

    QMutableListIterator<std::shared_ptr<ms::PromptSession>> iter(m_promptSessions);
    while(iter.hasNext()) {
        if (iter.next() == promptSession) {
            iter.remove();
            checkSessionChanges();
        }
    }
}

void Application::setSessionName(const QString& name)
{
    DLOG("Application::setSessionName (this=%p, name=%s)", this, name.toLatin1().constData());
    if (m_session) {
        LOG("Application::setSessionName should not be called once session exists");
        return;
    }
    m_sessionName = name;
}

void Application::setStage(Application::Stage stage)
{
    DLOG("Application::setStage (this=%p, stage=%d)", this, static_cast<int>(stage));
    if (m_stage != stage) {
        m_stage = stage;
        Q_EMIT stageChanged(stage);
    }
}

QImage Application::screenshotImage() const
{
    return m_screenshotImage;
}

void Application::updateScreenshot()
{
    if (!foregroundSession())
        return;
    QWeakPointer<Guard> wk(m_screenShotGuard.toWeakRef());

    foregroundSession()->take_snapshot(
        [&, wk](mir::scene::Snapshot const& snapshot)
        {
            // In case we get a threaded screenshot callback once the application is deleted.
            QMutexLocker lk(&screenshotMutex);
            if (wk.isNull())
                return;

            DLOG("ApplicationScreenshotProvider - Mir snapshot ready with size %d x %d",
                 snapshot.size.height.as_int(), snapshot.size.width.as_int());

            m_screenshotImage = QImage( (const uchar*)snapshot.pixels, // since we mirror, no need to offset starting position
                            snapshot.size.width.as_int(),
                            snapshot.size.height.as_int(),
                            QImage::Format_ARGB32_Premultiplied).mirrored();

            m_screenshot = QString("image://application/%1/%2").arg(m_desktopData->appId()).arg(QDateTime::currentMSecsSinceEpoch());
            Q_EMIT screenshotChanged(m_screenshot);
        });
}

void Application::setState(Application::State state)
{
    DLOG("Application::setState (this=%p, state=%d)", this, static_cast<int>(state));
    if (m_state != state) {
        switch (state)
        {
        case Application::Suspended:
            if (m_state == Application::Running) {
                if (session())
                    session()->set_lifecycle_state(mir_lifecycle_state_will_suspend);
                m_suspendTimer->start(3000);
            }
            break;
        case Application::Running:
            if (m_suspendTimer->isActive())
                m_suspendTimer->stop();

            if (m_state == Application::Suspended) {
                resume();
                if (session())
                    session()->set_lifecycle_state(mir_lifecycle_state_resumed);
            } else if (m_state == Application::Stopped) {
                respawn();
                state = Application::Starting;
            }
            break;
        case Application::Stopped:
            if (m_suspendTimer->isActive())
                m_suspendTimer->stop();
            break;
        default:
            break;
        }
        m_state = state;
        Q_EMIT stateChanged(state);
    }
}

void Application::setFocused(bool focused)
{
    DLOG("Application::setFocused (this=%p, focus=%s)", this, focused ? "yes" : "no");
    if (m_focused != focused) {
        m_focused = focused;
        Q_EMIT focusedChanged(focused);
    }
}

void Application::setFullscreen(bool fullscreen)
{
    DLOG("Application::setFullscreen (this=%p, fullscreen=%s)", this, fullscreen ? "yes" : "no");
    if (m_fullscreen != fullscreen) {
        m_fullscreen = fullscreen;
        Q_EMIT fullscreenChanged();
    }
}

void Application::setVisible(const bool visible)
{
    DLOG("Application::setVisible (this=%p, visible=%s)", this, visible ? "yes" : "no");
    // FIXME: this is bad, as should a MirSurface of this app exist, it won't be notified of the visiblity change.
    if (m_session && visible != m_visible) {
        if (visible) {
            m_session->show();
        } else {
            m_session->hide();
        }
        m_visible = visible;
        Q_EMIT visibleChanged();
    }
}

void Application::suspend()
{
    DLOG("Application::suspend (this=%p)", this);
    m_taskController->suspend(longAppId());
}

void Application::resume()
{
    DLOG("Application::resume (this=%p)", this);
    m_taskController->resume(longAppId());
}

void Application::respawn()
{
    DLOG("Application::respawn (this=%p)", this);
    m_taskController->start(appId(), m_arguments);
}

QString Application::longAppId() const
{
    return m_longAppId;
}

void Application::checkSessionChanges()
{
    ms::Session* lastTopSession = m_lastForegroundSession;
    ms::Session* currentTopSession = foregroundSession().get();
    m_lastForegroundSession = currentTopSession;
    if (currentTopSession && currentTopSession != lastTopSession) {
        Q_EMIT foregroundSessionChanged();
    }
}

} // namespace unitymir
