/*
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
 * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// local
#include "application_manager.h"
#include "proc_info.h"
#include "application.h"
#include "desktopfilereader.h"
#include "dbuswindowstack.h"
#include "taskcontroller.h"
#include "upstart/applicationcontroller.h"

// unity-mir
#include "qmirserverapplication.h"
#include "shellserverconfiguration.h"
#include "sessionlistener.h"
#include "sessionauthorizer.h"
#include "initialsurfaceplacementstrategy.h"
#include "taskcontroller.h"
#include "logging.h"
#include "promptsessionlistener.h"

// mir
#include <mir/scene/depth_id.h>
#include <mir/scene/session.h>
#include <mir/shell/focus_controller.h>
#include <mir/scene/surface.h>
#include <mir/scene/prompt_session.h>
#include <mir/graphics/display.h>
#include <mir/graphics/display_buffer.h>
#include <mircommon/mir/geometry/rectangles.h>
#include <mir/frontend/shell.h>
#include <mir/scene/prompt_session_manager.h>

// Qt
#include <QCoreApplication>
#include <QThread>

// std
#include <csignal>

// Default element sizes
#define SIDE_STAGE_WIDTH_GU 40
#define TABLET_MODE_MINIMUM_WIDTH_GU 100

namespace ms = mir::scene;

using namespace unity::shell::application;

namespace unitymir
{

namespace
{

QSize get_display_size(const std::shared_ptr<mir::graphics::Display> & display) {
    // Obtain display size
    mir::geometry::Rectangles view_area;
    display->for_each_display_buffer(
        [&view_area](const mir::graphics::DisplayBuffer & db)
        {
            view_area.add(db.view_area());
        });

    return QSize(
        view_area.bounding_rectangle().size.width.as_uint32_t(),
        view_area.bounding_rectangle().size.height.as_uint32_t()
        );
}

// FIXME: To be removed once shell has fully adopted short appIds!!
QString toShortAppIdIfPossible(const QString &appId) {
    QRegExp longAppIdMask("[a-z0-9][a-z0-9+.-]+_[a-zA-Z0-9+.-]+_[0-9][a-zA-Z0-9.+:~-]*");
    if (longAppIdMask.exactMatch(appId)) {
        LOG("WARNING: long App ID encountered: %s", qPrintable(appId));
        // input string a long AppId, chop the version string off the end
        QStringList parts = appId.split("_");
        return QString("%1_%2").arg(parts.first()).arg(parts.at(1));
    }
    return appId;
}

void connectToSessionListener(ApplicationManager * manager, SessionListener * listener)
{
    QObject::connect(listener, &SessionListener::sessionStarting,
                     manager, &ApplicationManager::onSessionStarting);
    QObject::connect(listener, &SessionListener::sessionStopping,
                     manager, &ApplicationManager::onSessionStopping);
    QObject::connect(listener, &SessionListener::sessionFocused,
                     manager, &ApplicationManager::onSessionFocused, Qt::QueuedConnection);
    QObject::connect(listener, &SessionListener::sessionUnfocused,
                     manager, &ApplicationManager::onSessionUnfocused);
    QObject::connect(listener, &SessionListener::sessionCreatedSurface,
                     manager, &ApplicationManager::onSessionCreatedSurface);
}

void connectToPromptSessionListener(ApplicationManager * manager, PromptSessionListener * listener)
{
    QObject::connect(listener, &PromptSessionListener::promptSessionStarting,
                     manager, &ApplicationManager::onPromptSessionStarting);
    QObject::connect(listener, &PromptSessionListener::promptSessionStopping,
                     manager, &ApplicationManager::onPromptSessionStopping);

    QObject::connect(listener, &PromptSessionListener::promptProviderAdded,
                     manager, &ApplicationManager::onPromptProviderAdded);
    QObject::connect(listener, &PromptSessionListener::promptProviderRemoved,
                     manager, &ApplicationManager::onPromptProviderRemoved);

}

void connectToSessionAuthorizer(ApplicationManager * manager, SessionAuthorizer * authorizer)
{
    QObject::connect(authorizer, &SessionAuthorizer::requestAuthorizationForSession,
                     manager, &ApplicationManager::authorizeSession, Qt::BlockingQueuedConnection);
}


void connectToPlacementStrategy(ApplicationManager * manager, InitialSurfacePlacementStrategy * strategy)
{
    QObject::connect(strategy, &InitialSurfacePlacementStrategy::requestPlacementForSession,
                     manager, &ApplicationManager::placeSession, Qt::DirectConnection);
}

void connectToTaskController(ApplicationManager * manager, TaskController * controller)
{
    QObject::connect(controller, &TaskController::processStarting,
                     manager, &ApplicationManager::onProcessStarting);
    QObject::connect(controller, &TaskController::processStopped,
                     manager, &ApplicationManager::onProcessStopped);
    QObject::connect(controller, &TaskController::processFailed,
                     manager, &ApplicationManager::onProcessFailed);
    QObject::connect(controller, &TaskController::requestFocus,
                     manager, &ApplicationManager::onFocusRequested);
    QObject::connect(controller, &TaskController::requestResume,
                     manager, &ApplicationManager::onResumeRequested);
}
} // namespace

ApplicationManager* ApplicationManager::Factory::Factory::create()
{
    QMirServerApplication* mirServerApplication = dynamic_cast<QMirServerApplication*>(QCoreApplication::instance());
    if (mirServerApplication == nullptr) {
        LOG("Need to use QMirServerApplication");
        QCoreApplication::quit();
        return nullptr;
    }

    ShellServerConfiguration * mirServer = mirServerApplication->server();

    QSize displaySize{get_display_size(mirServer->the_display())};

    QSharedPointer<upstart::ApplicationController> appController(new upstart::ApplicationController());
    QSharedPointer<TaskController> taskController(new TaskController(nullptr, appController));
    QSharedPointer<DesktopFileReader::Factory> fileReaderFactory(new DesktopFileReader::Factory());
    QSharedPointer<ProcInfo> procInfo(new ProcInfo());

    // FIXME: We should use a QSharedPointer to wrap this ApplicationManager object, which requires us
    // to use the data() method to pass the raw pointer to the QML engine. However the QML engine appears
    // to take ownership of the object, and deletes it when it wants to. This conflicts with the purpose
    // of the QSharedPointer, and a double-delete results. Trying QQmlEngine::setObjectOwnership on the
    // object no effect, which it should. Need to investigate why.
    ApplicationManager* appManager = new ApplicationManager(
                                             taskController,
                                             fileReaderFactory,
                                             procInfo,
                                             mirServer->the_focus_controller(),
                                             mirServer->the_prompt_session_manager(),
                                             displaySize
                                         );

    connectToSessionListener(appManager, mirServer->sessionListener());
    connectToPromptSessionListener(appManager, mirServer->promptSessionListener());
    connectToSessionAuthorizer(appManager, mirServer->sessionAuthorizer());
    connectToPlacementStrategy(appManager, mirServer->placementStrategy());
    connectToTaskController(appManager, taskController.data());

    return appManager;
}

ApplicationManager* ApplicationManager::singleton()
{
    static ApplicationManager* instance;
    if (!instance) {
        Factory appFactory;
        instance = appFactory.create();
    }
    return instance;
}

ApplicationManager::ApplicationManager(
        const QSharedPointer<TaskController>& taskController,
        const QSharedPointer<DesktopFileReader::Factory>& desktopFileReaderFactory,
        const QSharedPointer<ProcInfo>& procInfo,
        const std::shared_ptr<mir::shell::FocusController>& focusController,
        const std::shared_ptr<mir::scene::PromptSessionManager>& promptSessionManager,
        const QSize& displaySize,
        QObject *parent)
    : ApplicationManagerInterface(parent)
    , m_focusedApplication(nullptr)
    , m_mainStageApplication(nullptr)
    , m_sideStageApplication(nullptr)
    , m_msApplicationToBeFocused(nullptr)
    , m_ssApplicationToBeFocused(nullptr)
    , m_lifecycleExceptions(QStringList() << "com.ubuntu.music")
    , m_focusController(focusController)
    , m_promptSessionManager(promptSessionManager)
    , m_dbusWindowStack(new DBusWindowStack(this))
    , m_taskController(taskController)
    , m_desktopFileReaderFactory(desktopFileReaderFactory)
    , m_procInfo(procInfo)
    , m_gridUnitPx(8)
    , m_displaySize(displaySize)
    , m_panelHeight(54)
{
    DLOG("ApplicationManager::ApplicationManager (this=%p)", this);

    // Setup panel height
    QByteArray gridUnitString = qgetenv("GRID_UNIT_PX");
    if (!gridUnitString.isEmpty()) {
        bool ok;
        int value = gridUnitString.toInt(&ok);
        if (ok) {
            m_gridUnitPx = value;
        }
    }

    int densityPixelPx = qFloor( (float)m_gridUnitPx / 8 );

    m_panelHeight = 3 * m_gridUnitPx + 2 * densityPixelPx;
}

ApplicationManager::~ApplicationManager()
{
    DLOG("ApplicationManager::~ApplicationManager");
}

int ApplicationManager::panelHeight()
{
    return m_panelHeight;
}

int ApplicationManager::rowCount(const QModelIndex &parent) const
{
    return !parent.isValid() ? m_applications.size() : 0;
}

QVariant ApplicationManager::data(const QModelIndex &index, int role) const
{
    if (index.row() >= 0 && index.row() < m_applications.size()) {
        Application *application = m_applications.at(index.row());
        switch (role) {
            case RoleAppId:
                return QVariant::fromValue(application->appId());
            case RoleName:
                return QVariant::fromValue(application->name());
            case RoleComment:
                return QVariant::fromValue(application->comment());
            case RoleIcon:
                return QVariant::fromValue(application->icon());
            case RoleStage:
                return QVariant::fromValue((int)application->stage());
            case RoleState:
                return QVariant::fromValue((int)application->state());
            case RoleFocused:
                return QVariant::fromValue(application->focused());
            case RoleScreenshot:
                return QVariant::fromValue(application->screenshot());
            default:
                return QVariant();
        }
    } else {
        return QVariant();
    }
}

Application* ApplicationManager::get(int index) const
{
    DLOG("ApplicationManager::get (this=%p, index=%i, count=%i)", this, index, m_applications.count());
    if (index < 0 || index >= m_applications.count()) {
        return nullptr;
    }
    return m_applications.at(index);
}

Application* ApplicationManager::findApplication(const QString &inputAppId) const
{
    const QString appId = toShortAppIdIfPossible(inputAppId);

    for (Application *app : m_applications) {
        if (app->appId() == appId) {
            return app;
        }
    }
    return nullptr;
}

bool ApplicationManager::requestFocusApplication(const QString &inputAppId)
{
    const QString appId = toShortAppIdIfPossible(inputAppId);

    DLOG("ApplicationManager::requestFocusApplication (this=%p, appId=%s)", this, qPrintable(appId));
    Application *application = findApplication(appId);

    if (!application) {
        DLOG("No such running application '%s'", qPrintable(appId));
        return false;
    }

    if (application == m_focusedApplication) {
        DLOG("Application %s is already focused", qPrintable(appId));
        return false;
    }

    // If there is a currently focused application, first update the screenshot
    // for it and only emit focusRequested() when the screenshot is ready.
    Application *currentlyFocusedApplication = findApplication(focusedApplicationId());
    if (currentlyFocusedApplication) {
        m_nextFocusedAppId = appId;
        currentlyFocusedApplication->updateScreenshot();
    } else {
        Q_EMIT focusRequested(appId);
    }
    return true;
}

QString ApplicationManager::focusedApplicationId() const
{
    if (m_focusedApplication) {
        return m_focusedApplication->appId();
    } else {
        return QString();
    }
}

bool ApplicationManager::suspended() const
{
    return m_suspended;
}

void ApplicationManager::setSuspended(bool suspended)
{
    if (suspended == m_suspended) {
        return;
    }
    m_suspended = suspended;
    Q_EMIT suspendedChanged();

    if (m_suspended) {
        suspendApplication(m_mainStageApplication);
        suspendApplication(m_sideStageApplication);
    } else {
        resumeApplication(m_mainStageApplication);
        resumeApplication(m_sideStageApplication);
    }
}

bool ApplicationManager::suspendApplication(Application *application)
{
    if (application == nullptr)
        return false;

    updateScreenshot(application->appId());

    DLOG("ApplicationManager::suspend(this=%p, application(%p)->appId(%s) )",this, application, qPrintable(application->appId()));
    // Present in exceptions list, return.
    if (!m_lifecycleExceptions.filter(application->appId().section('_',0,0)).empty())
        return false;

    if (application->state() == Application::Running)
        application->setState(Application::Suspended);

    return true;
}

void ApplicationManager::resumeApplication(Application *application)
{
    if (application == nullptr)
        return;

    if (application->state() != Application::Running) {
        application->setVisible(true);
        application->setState(Application::Running);
    }
}

bool ApplicationManager::focusApplication(const QString &inputAppId)
{
    const QString appId = toShortAppIdIfPossible(inputAppId);

    Application *application = findApplication(appId);
    DLOG("ApplicationManager::focusApplication (this=%p, application=%p, appId=%s)", this, application, qPrintable(appId));

    if (!application) {
        DLOG("No such running application '%s'", qPrintable(appId));
        return false;
    }

    if (application->stage() == Application::MainStage && m_sideStageApplication)
        suspendApplication(m_sideStageApplication);

    if (application->stage() == Application::MainStage)
        m_msApplicationToBeFocused = application;
    else
        m_ssApplicationToBeFocused = application;

    if (application->state() == Application::Stopped) {
        // Respawning this app, move to end of application list so onSessionStarting works ok
        // FIXME: this happens pretty late, shell could request respawn earlier
        application->setState(Application::Running);
        int from = m_applications.indexOf(application);
        move(from, m_applications.length()-1);
    } else {
        auto mirSession = application->foregroundSession();
        if (mirSession) {
            m_focusController->set_focus_to(mirSession);
            int from = m_applications.indexOf(application);
            move(from, 0);
        }
    }

    // FIXME(dandrader): lying here. The operation is async. So we will only know whether
    // the focusing was successful once the server replies. Maybe the API in unity-api should
    // reflect that?
    return true;
}

void ApplicationManager::unfocusCurrentApplication()
{
    DLOG("ApplicationManager::unfocusCurrentApplication (this=%p)", this);

    suspendApplication(m_sideStageApplication);
    suspendApplication(m_mainStageApplication);

    // Clear both stages
    m_msApplicationToBeFocused = nullptr;
    m_ssApplicationToBeFocused = nullptr;
    m_focusController->set_focus_to(nullptr); //FIXME(greyback)
}

Application* ApplicationManager::startApplication(const QString &appId,
                                                  const QStringList &arguments)
{
    return startApplication(appId, NoFlag, arguments);
}

/**
 * @brief ApplicationManager::startApplication launches an application identified by an "application id" or appId.
 *
 * Note: due to an implementation detail, appIds come in two forms:
 * * long appId: $(click_package)_$(application)_$(version)
 * * short appId: $(click_package)_$(application)
 * It is expected that the shell uses _only_ short appIds (but long appIds are accepted by this method for legacy
 * reasons - but be warned, this ability will be removed)
 *
 * Unless stated otherwise, we always use short appIds in this API.
 *
 * @param inputAppId AppId of application to launch (long appId supported)
 * @param flags Shell specific flags to modify the initial state of the application
 * @param arguments Command line arguments to pass to the application to be launched
 * @return Pointer to Application object representing the launched process
 */
Application *ApplicationManager::startApplication(const QString &inputAppId, ExecFlags flags,
                                                  const QStringList &arguments)
{
    QString appId = toShortAppIdIfPossible(inputAppId);
    DLOG("ApplicationManager::startApplication (this=%p, appId=%s)", this, qPrintable(appId));

    if (!m_taskController->start(appId, arguments)) {
        LOG("Asking Upstart to start application '%s' failed", qPrintable(appId));
        return nullptr;
    }

    {
        Application * application = findApplication(appId);
        if (application) {
            DLOG("ApplicationManager::startApplication - application already exists: (this=%p, app=%p, appId=%s)",
                 this, application, qPrintable(appId));
            return application;
        }
    }

    Application* application = new Application(
                m_taskController,
                m_desktopFileReaderFactory->createInstance(appId, m_taskController->findDesktopFileForAppId(appId)),
                Application::Starting,
                arguments,
                m_promptSessionManager,
                this);
    if (!application->isValid()) {
        DLOG("Unable to instantiate application with appId '%s'", qPrintable(appId));
        return nullptr;
    }

    // override stage if necessary
    if (application->stage() == Application::SideStage && flags.testFlag(ApplicationManager::ForceMainStage)) {
        application->setStage(Application::MainStage);
    }

    add(application);
    return application;
}

void ApplicationManager::onProcessStarting(const QString &appId)
{
    DLOG("ApplicationManager::onProcessStarting (this=%p, appId=%s)", this, qPrintable(appId));
    Application *application = findApplication(appId);

    if (!application) { // then shell did not start this application, so upstart must have - add to list
        application = new Application(
                    m_taskController,
                    m_desktopFileReaderFactory->createInstance(appId, m_taskController->findDesktopFileForAppId(appId)),
                    Application::Starting,
                    QStringList(),
                    m_promptSessionManager,
                    this);
        if (!application->isValid()) {
            DLOG("Unable to instantiate application with appId '%s'", qPrintable(appId));
            return;
        }

        // override stage if necessary (i.e. side stage invalid on phone)
        if (application->stage() == Application::SideStage &&
                m_displaySize.width() < (TABLET_MODE_MINIMUM_WIDTH_GU * m_gridUnitPx)) {
            application->setStage(Application::MainStage);
        }

        add(application);
        Q_EMIT focusRequested(appId);
    }
    else {
        DLOG("ApplicationManager::onProcessStarting application already found: (app=%p, appId=%s)", application, qPrintable(appId));
    }
}

void ApplicationManager::onProcessFailed(const QString &appId, const bool duringStartup)
{
    /* Applications fail if they fail to launch, crash or are killed. If failed to start, must
     * immediately remove from list of applications. If crash or kill, instead we set flag on the
     * Application to indicate it can be resumed.
     */

    DLOG("ApplicationManager::onProcessFailed (this=%p, appId=%s, duringStartup=%c)",
         this, qPrintable(appId), (duringStartup) ? 'Y' : 'N');

    Application *application = findApplication(appId);
    if (!application) {
        LOG("ApplicationManager::onProcessFailed - upstart reports failure of application AppManager is not managing");
        return;
    }

    Q_UNUSED(duringStartup); // FIXME(greyback) upstart reports app that fully started up & crashes as failing during startup??
    if (application->state() == Application::Starting) {
        checkFocusOnRemovedApplication(application);
        remove(application);
        m_dbusWindowStack->WindowDestroyed(0, application->appId());
        delete application;
    } else {
        // We need to set flags on the Application to say the app can be resumed, and thus should not be removed
        // from the list by onProcessStopped.
        application->setCanBeResumed(true);
        application->setPid(0);
    }
}

bool ApplicationManager::stopApplication(const QString &inputAppId)
{
    const QString appId = toShortAppIdIfPossible(inputAppId);

    Application *application = findApplication(appId);
    DLOG("ApplicationManager::stopApplication (this=%p, application=%p, appId=%s)", this, application, qPrintable(appId));

    if (!application) {
        DLOG("No such running application '%s'", qPrintable(appId));
        return false;
    }

    checkFocusOnRemovedApplication(application);

    remove(application);
    m_dbusWindowStack->WindowDestroyed(0, application->appId());

    bool result = m_taskController->stop(application->longAppId());

    LOG_IF(result == false, "FAILED to ask Upstart to stop application '%s'", qPrintable(application->appId()));

    if (!result && application->pid() > 0) {
        DLOG("TaskController::stop UAL failed. Sending SIGTERM to %i", application->pid());
        kill(application->pid(), SIGTERM);
    }

    delete application;

    // FIXME(dandrader): lying here. The operation is async. So we will only know whether
    // the focusing was successful once the server replies. Maybe the API in unity-api should
    // reflect that?
    return result;
}

bool ApplicationManager::updateScreenshot(const QString &inputAppId)
{
    const QString appId = toShortAppIdIfPossible(inputAppId);

    Application *application = findApplication(appId);

    if (!application) {
        DLOG("No such running application '%s'", qPrintable(appId));
        return false;
    }

    application->updateScreenshot();
    QModelIndex appIndex = findIndex(application);
    Q_EMIT dataChanged(appIndex, appIndex, QVector<int>() << RoleScreenshot);
    return true;
}

void ApplicationManager::onProcessStopped(const QString &appId)
{
    Application *application = findApplication(appId);
    if (!application) {
        DLOG("ApplicationManager::onProcessStopped reports stop of appId='%s' which AppMan is not managing, ignoring the event");
        return;
    }
    DLOG("ApplicationManager::onProcessStopped (this=%p, application=%p, appId=%s)", this, application, qPrintable(appId));

    bool removeApplication = true;

    // The following scenario is the only time that we do NOT remove the application from the app list:
    if ((application->state() == Application::Suspended || application->state() == Application::Stopped)
            && application->pid() == 0 // i.e. onProcessFailed was called, which resets the PID of this application
            && application->canBeResumed()) {
        removeApplication = false;
    }

    if (removeApplication) {
        DLOG("ApplicationManager::onProcessStopped - removing appId='%s' from the application list", qPrintable(appId));
        checkFocusOnRemovedApplication(application);
        remove(application);
        m_dbusWindowStack->WindowDestroyed(0, application->appId());
        delete application;
    }
}

void ApplicationManager::onFocusRequested(const QString& appId)
{
    DLOG("ApplicationManager::onFocusRequested (this=%p, appId=%s)", this, qPrintable(appId));

    Q_EMIT focusRequested(appId);
}

void ApplicationManager::onResumeRequested(const QString& appId)
{
    Application *application = findApplication(appId);
    DLOG("ApplicationManager::onResumeRequested (this=%p, application=%p, appId=%s)", this, application, qPrintable(appId));


    if (!application) {
        DLOG("ApplicationManager::onResumeRequested: No such running application '%s'", qPrintable(appId));
        return;
    }

    // If app Stopped, trust that upstart-app-launch respawns it itself, and AppManager will
    // be notified of that through the onProcessStarting slot. Else resume.
    if (application->state() == Application::Suspended) {
        application->setVisible(true);
        application->setState(Application::Running);
    }
}

void ApplicationManager::screenshotUpdated()
{
    if (sender()) {
        Application *application = static_cast<Application*>(sender());
        QModelIndex appIndex = findIndex(application);
        Q_EMIT dataChanged(appIndex, appIndex, QVector<int>() << RoleScreenshot);

        DLOG("ApplicationManager::screenshotUpdated: Received new screenshot for %s", qPrintable(application->appId()));
    } else {
        DLOG("ApplicationManager::screenshotUpdated: Received screenshotUpdated signal but application has disappeard.");
    }

    // Now that the screenshot is ready, check if there
    // is a pending app focusing request.
    if (!m_nextFocusedAppId.isEmpty()) {
        Q_EMIT focusRequested(m_nextFocusedAppId);
        m_nextFocusedAppId.clear();
    }
}

void ApplicationManager::authorizeSession(const quint64 pid, bool &authorized)
{
    authorized = false; //to be proven wrong

    DLOG("ApplicationManager::authorizeSession (this=%p, pid=%lld)", this, pid);

    for (Application *app : m_applications) {
        if (app->state() == Application::Starting
                && m_taskController->appIdHasProcessId(app->longAppId(), pid)) {
            app->setPid(pid);
            DLOG("ApplicationManager::authorizeSession - connecting: application=%p and pid=%lld", app, pid);
            authorized = true;
            return;
        }
    }

    /*
     * Hack: Allow applications to be launched externally, but must be executed with the
     * "desktop_file_hint" parameter attached. This exists until upstart-app-launch can
     * notify shell it is starting an application and so shell should allow it. Also reads
     * the --stage parameter to determine the desired stage
     */
    std::unique_ptr<ProcInfo::CommandLine> info = m_procInfo->commandLine(pid);
    if (!info) {
        DLOG("ApplicationManager REJECTED connection from app with pid %lld as unable to read process command", pid);
        return;
    }

    if (info->startsWith("maliit-server") || info->contains("qt5/libexec/QtWebProcess")) {
        authorized = true;
        m_hiddenPIDs << pid;
        return;
    }

    boost::optional<QString> desktopFileName{ info->getParameter("--desktop_file_hint=") };

    if (!desktopFileName) {
        LOG("ApplicationManager REJECTED connection from app with pid %lld as no desktop_file_hint specified (%s)", pid,
            qPrintable(info->asStringList().first()));
        return;
    }

    DLOG("Process supplied desktop_file_hint, loading '%s'", desktopFileName.get().toLatin1().data());

    // Guess appId from the desktop file hint
    QString appId = toShortAppIdIfPossible(desktopFileName.get().remove(QRegExp(".desktop$")).split('/').last());

    // FIXME: right now we support --desktop_file_hint=appId for historical reasons. So let's try that in
    // case we didn't get an existing .desktop file path
    DesktopFileReader* desktopData;
    if (QFileInfo(desktopFileName.get()).exists()) {
        desktopData = m_desktopFileReaderFactory->createInstance(appId, QFileInfo(desktopFileName.get()));
    } else {
        desktopData = m_desktopFileReaderFactory->createInstance(appId, m_taskController->findDesktopFileForAppId(appId));
    }

    if (!desktopData->loaded()) {
        delete desktopData;
        LOG("ApplicationManager REJECTED connection from app with pid %lld as desktop_file_hint file not found", pid);
        return;
    }

    // some naughty applications use a script to launch the actual application. Check for the
    // case where shell actually launched the script.
    Application *application = findApplication(desktopData->appId());
    if (application && application->state() == Application::Starting) {
        DLOG("Process with pid %lld appeared, attached to existing entry '%s' in application lists",
             pid, application->appId().toLatin1().data());
        delete desktopData;
        application->setSessionName(application->appId());
        application->setPid(pid);
        authorized = true;
        return;
    }

    // if stage supplied in CLI, fetch that
    Application::Stage stage = Application::MainStage;
    boost::optional<QString> stageParam = info->getParameter("--stage_hint=");

    if (stageParam && stageParam.get() == "side_stage") {
        stage = Application::SideStage;
    }

    DLOG("Existing process with pid %lld appeared, adding '%s' to application lists", pid, desktopData->name().toLatin1().data());

    QStringList arguments(info->asStringList());
    application = new Application(
        m_taskController,
        desktopData,
        Application::Starting,
        arguments,
        m_promptSessionManager,
        this);
    application->setPid(pid);
    application->setStage(stage);
    application->setCanBeResumed(false);
    add(application);
    authorized = true;
}

void ApplicationManager::placeSession(ms::Session const* session, uint32_t &x, uint32_t &y)
{
    Application* application = findApplicationWithSession(session, true);
    DLOG("ApplicationManager::placeSession (this=%p, application=%p, session=%p, name=%s)", this, application, session,
         session?(session->name().c_str()):"null");

    // Application defaults
    x = 0;
    y = m_panelHeight;

    // Shell client override
    if (!session) {
        y = 0;
        return;
    }

    // Fullscreen applications override
    if (application && application->fullscreen())
        y = 0;

    // SideStage override
    if (application && application->stage() == Application::SideStage)
        x = m_displaySize.width() - (SIDE_STAGE_WIDTH_GU * m_gridUnitPx);

    DLOG("ApplicationManager::placeSession (x=%d, y=%d)", x, y);
}

void ApplicationManager::onSessionStarting(const std::shared_ptr<ms::Session>& session)
{
    DLOG("ApplicationManager::onSessionStarting (this=%p, application=%s)", this, session?session->name().c_str():"null");

    pid_t pid = session->process_id();

    Application* application = findApplicationWithPid(session->process_id(), false);
    if (application && application->state() != Application::Running) {
        application->setSession(session);
        if (application->stage() == Application::MainStage)
            m_msApplicationToBeFocused = application;
        else
            m_ssApplicationToBeFocused = application;
    } else {
        if (m_hiddenPIDs.contains(pid)) {
            return;
        }
        DLOG("ApplicationManager::onSessionStarting - unauthorized application!!");
    }
}

void ApplicationManager::onSessionStopping(const std::shared_ptr<ms::Session>& session)
{
    DLOG("ApplicationManager::onSessionStopping (this=%p, session name=%s)", this, session?session->name().c_str():"null");

    Application* application = findApplicationWithSession(session, false); // only check app sessions
    if (application) {
        /* Can remove the application from the running apps list immediately in these curcumstances:
         *  1. application is not managed by upstart (this message from Mir is only notice the app has stopped, must do
         *     it here)
         *  2. application is managed by upstart, but has stopped before it managed to create a surface, we can assume
         *     it crashed on startup, and thus cannot be resumed - so remove it.
         *  3. application is managed by upstart and is in foreground (i.e. has Running state), if Mir reports the
         *     application disconnects, it either crashed or stopped itself. Either case, remove it.
         */
        if (!application->canBeResumed()
                || application->state() == Application::Starting
                || application->state() == Application::Running) {
            checkFocusOnRemovedApplication(application);
            m_dbusWindowStack->WindowDestroyed(0, application->appId());
            remove(application);
            delete application;
        } else {
            // otherwise, we do not have enough information to make any changes to the model, so await events from
            // upstart to go further, but set the app state
            application->setState(Application::Stopped);
        }
    }

    m_hiddenPIDs.removeOne(session->process_id());
}

void ApplicationManager::onSessionFocused(const std::shared_ptr<ms::Session>& session)
{
    Application* application = findApplicationWithSession(session, true);
    DLOG("ApplicationManager::onSessionFocused (this=%p, application=%p, appId=%s, session name=%s)", this, application,
         application?qPrintable(application->appId()):"null", session?session->name().c_str():"null");

    // Don't give application focus until it has created it's surface, when it is set as state "Running"
    // and only notify shell of focus changes that it actually expects
    if (application && application->state() != Application::Starting &&
        (application == m_msApplicationToBeFocused ||
         application == m_ssApplicationToBeFocused)
            && application != m_focusedApplication) {
        setFocused(application);
        QModelIndex appIndex = findIndex(application);
        Q_EMIT dataChanged(appIndex, appIndex, QVector<int>() << RoleFocused);
    } else {
        if (application == nullptr) {
            DLOG("Invalid application focused, discarding the event");
            if (nullptr != m_focusedApplication) {
                focusApplication(m_focusedApplication->appId());
            } else {
                m_focusController->set_focus_to(nullptr);
            }
        } else if (application == m_focusedApplication && session != application->foregroundSession()) {
            DLOG("Focused session is not on top. Resetting focus");
            focusApplication(application->appId());
        }
    }
}

void ApplicationManager::onSessionUnfocused()
{
    DLOG("ApplicationManager::onSessionUnfocused (this=%p, application=%p)", this, m_focusedApplication);
    if (nullptr != m_focusedApplication) {
        Q_ASSERT(m_focusedApplication->focused());
        m_focusedApplication->setFocused(false);

        auto lastFocusedApp = m_focusedApplication;
        suspendApplication(m_focusedApplication);

        m_focusedApplication = nullptr;
        Q_EMIT focusedApplicationIdChanged();
        m_dbusWindowStack->FocusedWindowChanged(0, QString(), 0);

        QModelIndex appIndex = findIndex(m_focusedApplication);
        Q_EMIT dataChanged(appIndex, appIndex, QVector<int>() << RoleFocused << RoleState);

        if (lastFocusedApp && lastFocusedApp->activePromptSession()) {
            m_promptSessionManager->stop_prompt_session(lastFocusedApp->activePromptSession());
        }
    }
}

void ApplicationManager::onSessionCreatedSurface(const ms::Session * session,
                                                 const std::shared_ptr<ms::Surface> & surface)
{
    DLOG("ApplicationManager::onSessionCreatedSurface (this=%p, appId=%s)", this, session->name().c_str());
    Q_UNUSED(surface);

    Application* application = findApplicationWithSession(session, true);
    if (application) {
        if (application->state() == Application::Starting) {
            m_dbusWindowStack->WindowCreated(0, application->appId());
            // only when Session creates a Surface will we actually mark it focused
            setFocused(application);
            QModelIndex appIndex = findIndex(application);
            Q_EMIT dataChanged(appIndex, appIndex, QVector<int>() << RoleFocused);
        }

        application->checkSessionChanges();
    }
}

void ApplicationManager::onPromptSessionStarting(const std::shared_ptr<ms::PromptSession>& promptSession)
{
    DLOG("ApplicationManager::onPromptSessionStarting (this=%p, prompt_session=%p)", this, promptSession.get());

    Application* app = findApplicationWithSession(m_promptSessionManager->application_for(promptSession), false);
    if (app) {
        app->appendPromptSession(promptSession);
    } else {
        DLOG("ApplicationManager::onPromptSessionStarting - no app found (prompt_session=%p)", promptSession.get());
    }
}

void ApplicationManager::onPromptSessionStopping(const std::shared_ptr<ms::PromptSession>& promptSession)
{
    DLOG("ApplicationManager::onPromptSessionStopping (this=%p, prompt_session=%p)", this, promptSession.get());

    for (Application *app : m_applications) {
        app->removePromptSession(promptSession);
    }
}

void ApplicationManager::onPromptProviderAdded(ms::PromptSession const* promptSession,
                                               const std::shared_ptr<ms::Session>& session)
{
    Q_UNUSED(promptSession);
    DLOG("ApplicationManager::onPromptProviderAdded (this=%p, prompt_session=%p, session=%s)", this, promptSession, session?session->name().c_str():"null");

    Application* app = findApplicationWithSession(session, true);
    if (app) {
        app->checkSessionChanges();

        if (app == m_focusedApplication) {
            focusApplication(app->appId());
        }
    } else {
        DLOG("ApplicationManager::onPromptProviderAdded - no app found (prompt_session=%p)", promptSession);
    }
}

void ApplicationManager::onPromptProviderRemoved(ms::PromptSession const* promptSession,
                                                 const std::shared_ptr<ms::Session>& session)
{
    Q_UNUSED(session);
    DLOG("ApplicationManager::onPromptProviderAdded (this=%p, prompt_session=%p, session=%s)", this, (void*)promptSession, session?session->name().c_str():"null");

    // we need to update the focus if the app is in focus.
    for (Application *app : m_applications) {
        if (promptSession == app->activePromptSession().get()) {
            app->checkSessionChanges();

            if (app == m_focusedApplication) {
                focusApplication(app->appId());
            }
        }
    }
}

void ApplicationManager::setFocused(Application *application)
{
    DLOG("ApplicationManager::setFocused (application=%p, appId=%s)", application, qPrintable(application->appId()));

    if (application == m_focusedApplication)
        return;

    // set state of previously focused app to suspended
    if (m_focusedApplication && m_lifecycleExceptions.filter(m_focusedApplication->appId().section('_',0,0)).empty()) {
        Application *lastApplication = applicationForStage(application->stage());
        if (suspendApplication(lastApplication))
            lastApplication->setVisible(false);
    }

    if (application->stage() == Application::MainStage)
        m_mainStageApplication = application;
    else
        m_sideStageApplication = application;

    auto lastFocusedApp = m_focusedApplication;
    m_focusedApplication = application;
    m_focusedApplication->setFocused(true);
    m_focusedApplication->setVisible(true);
    m_focusedApplication->setState(Application::Running);
    move(m_applications.indexOf(application), 0);
    Q_EMIT focusedApplicationIdChanged();
    m_dbusWindowStack->FocusedWindowChanged(0, application->appId(), application->stage());

    // stop prompt session when focus changes.
    if (lastFocusedApp && lastFocusedApp->activePromptSession()) {
        m_promptSessionManager->stop_prompt_session(lastFocusedApp->activePromptSession());
    }
}

Application* ApplicationManager::findApplicationWithSession(const std::shared_ptr<ms::Session> &session, bool includeChildSessions)
{
    return findApplicationWithSession(session.get(), includeChildSessions);
}

Application* ApplicationManager::findApplicationWithSession(const ms::Session *session, bool includeChildSessions)
{
    if (!session)
        return nullptr;
    return findApplicationWithPid(session->process_id(), includeChildSessions);
}

Application* ApplicationManager::findApplicationWithPid(const qint64 pid, bool includeChildSessions)
{
    if (pid <= 0)
        return nullptr;

    for (Application *app : m_applications) {
        if (app->m_pid == pid) {
            return app;
        }

        if (includeChildSessions && app->containsProcess(pid)) {
            return app;
        }
    }
    return nullptr;
}

Application* ApplicationManager::applicationForStage(Application::Stage stage)
{
    DLOG("ApplicationManager::focusedApplicationForStage(this=%p)", this);

    if (stage == Application::MainStage)
        return m_mainStageApplication;
    else
        return m_sideStageApplication;
}

void ApplicationManager::add(Application* application)
{
    DASSERT(application != nullptr);
    DLOG("ApplicationManager::add (this=%p, application='%s')", this, qPrintable(application->name()));

    connect(application, &Application::screenshotChanged, this, &ApplicationManager::screenshotUpdated);

    beginInsertRows(QModelIndex(), m_applications.size(), m_applications.size());
    m_applications.append(application);
    endInsertRows();
    Q_EMIT countChanged();
    Q_EMIT applicationAdded(application->appId());
}

void ApplicationManager::remove(Application *application)
{
    DASSERT(application != nullptr);
    DLOG("ApplicationManager::remove (this=%p, application='%s')", this, qPrintable(application->name()));

    if (application == m_sideStageApplication)
        m_sideStageApplication = nullptr;
    if (application == m_mainStageApplication)
        m_mainStageApplication = nullptr;

    int i = m_applications.indexOf(application);
    if (i != -1) {
        beginRemoveRows(QModelIndex(), i, i);
        m_applications.removeAt(i);
        endRemoveRows();
        Q_EMIT applicationRemoved(application->appId());
        Q_EMIT countChanged();
    }
}

void ApplicationManager::move(int from, int to)
{
    DLOG("ApplicationManager::move (this=%p, from=%d, to=%d)", this, from, to);
    if (from == to) return;

    if (from >= 0 && from < m_applications.size() && to >= 0 && to < m_applications.size()) {
        QModelIndex parent;
        /* When moving an item down, the destination index needs to be incremented
           by one, as explained in the documentation:
           http://qt-project.org/doc/qt-5.0/qtcore/qabstractitemmodel.html#beginMoveRows */
        beginMoveRows(parent, from, from, parent, to + (to > from ? 1 : 0));
        m_applications.move(from, to);
        endMoveRows();
    }
}

QModelIndex ApplicationManager::findIndex(Application* application)
{
    for (int i = 0; i < m_applications.size(); ++i) {
        if (m_applications.at(i) == application) {
            return index(i);
        }
    }

    return QModelIndex();
}

bool ApplicationManager::checkFocusOnRemovedApplication(Application * application)
{
    if (application == m_focusedApplication) {
        // TODO(greyback) What to do?? Focus next app, or unfocus everything??
        auto lastFocusedApp = m_focusedApplication;
        m_focusedApplication = nullptr;
        Q_EMIT focusedApplicationIdChanged();

        // stop prompt session when focus changes.
        if (lastFocusedApp && lastFocusedApp->activePromptSession()) {
            m_promptSessionManager->stop_prompt_session(lastFocusedApp->activePromptSession());
        }

        return true;
    }
    return false;
}

} // namespace unitymir
